"""Flask test client extensions.
"""

from flask.testing import FlaskClient
from flask import json
from werkzeug import urls
from werkzeug.utils import cached_property


class JsonResponseMixin(object):
    """Mixin which adds method to jsonify response data"""
    @cached_property
    def json(self):
        """Attempt to convert response data to JSON."""
        return json.loads(self.data)


def make_client_response(response_class):
    """Factory function for mixing JsonResponseMixin with app's response class.
    """
    class ClientResponse(response_class, JsonResponseMixin):
        """Mixed in resposne class."""
        pass

    return ClientResponse


class Client(FlaskClient):
    """Expose standard interface for HTTP verbs.
    """
    def get(self, url, params=None, **kargs):
        """Wrap get request with URL params support."""
        if params:
            url = urls.Href(url)(params)

        return super(Client, self).get(url, **kargs)

    def post(self, url, data, **kargs):
        """Wrap post request with data argument exposed in args."""
        kargs['data'] = data
        return super(Client, self).post(url, **kargs)

    def put(self, url, data, **kargs):
        """Wrap put request with data argument exposed in args."""
        kargs['data'] = data
        return super(Client, self).put(url, **kargs)

    def patch(self, url, data, **kargs):
        """Wrap patch request with data argument exposed in args."""
        kargs['data'] = data
        return super(Client, self).patch(url, **kargs)


class JSONClient(Client):
    """JSON API client with convenience handling of content-type and JSON
    serialization.
    """
    def open(self, *args, **kargs):
        """Wrap open with JSON data handling."""

        # All requests will be treated like JSON unless otherwise specified.
        kargs.setdefault('content_type', 'application/json')

        if (kargs['content_type'] == 'application/json'
                and isinstance(kargs.get('data'), dict)):
            # If data is a dict, then assume we want to send a JSON serialized
            # string in the request.
            try:
                kargs['data'] = json.dumps(kargs['data'])
            except Exception:
                # Ignore error if data isn't serializable and just send it.
                pass

        return super(JSONClient, self).open(*args, **kargs)



"""Extension of flask.Request.
"""

from flask import Request as RequestBase


class Request(RequestBase):
    """Subclass of flask.Request with some added features"""

    @property
    def data(self):
        """Property access to get_dict()."""
        return self.get_dict()

    def get_dict(self, force=True, silent=True, cache=True):
        """Attempt to return request data as a dict. This is similar to
        `get_json` but is more permissive in trying to return something useful.
        Try to convert from JSON first but then fallback to other extraction
        methods.
        """
        data = getattr(self, '_cached_dict', None)
        if data is not None:
            return data

        data = self.get_json(force=force, silent=silent, cache=cache)

        if data is None:
            # fallback to form data
            data = self.form

        if hasattr(data, 'to_dict'):
            data = data.to_dict()

        data = data or {}

        if cache:
            self._cached_dict = data

        return data



"""Session interface for Flask.
"""

from flask.sessions import SecureCookieSessionInterface


class SessionInterface(SecureCookieSessionInterface):
    """Extends secure cookie session interface with better support for
    permanent sessions."""
    def __init__(self, salt=None, permanent=True):
        if salt:
            # Salt is added on top of secret key.
            self.salt = salt

        self.permanent = permanent

    def open_session(self, app, request):
        session = super(SessionInterface, self).open_session(app, request)

        if session is not None and self.permanent:
            session.permanent = True

        return session


def init_app(app):
    """Initialize app."""
    app.config.setdefault('CARAFE_SESSION_ENABLED', True)
    app.config.setdefault('CARAFE_SESSION_PERMANENT', True)
    app.config.setdefault('CARAFE_SESSION_SALT', None)

    if not app.config['CARAFE_SESSION_ENABLED']:
        return

    permanent = (app.config['PERMANENT_SESSION_LIFETIME'].total_seconds() > 0
                 and app.config['CARAFE_SESSION_PERMANENT'])

    app.session_interface = SessionInterface(
        salt=app.config['CARAFE_SESSION_SALT'], permanent=permanent)



"""General purpose utility functions.
"""

from functools import wraps, partial
from threading import Thread

from flask import jsonify as _jsonify


class classproperty(object):  # pylint: disable=invalid-name
    """Decorator that adds class properties. Allows for usage like @property
    but applies the property at the class level. Helps avoid having to use
    metaclasses or other complex techniques to achieve similar results.
    """
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def iterflatten(items):
    """Return iterator which flattens list/tuple of lists/tuples

    >>> lst = [1, [2,3], [4, [5, [6]], 7], 8]
    >>> assert list(iterflatten(lst)) == [1,2,3,4,5,6,7,8]
    """
    for item in items:
        if isinstance(item, (list, tuple)):
            for subitem in flatten(item):
                yield subitem
        else:
            yield item


def flatten(items):
    """Return flattened list of a list/tuple of lists/tuples

    >>> lst = [1, [2,3], [4, [5, [6]], 7], 8]
    >>> assert flatten(lst) == [1,2,3,4,5,6,7,8]
    """
    return list(iterflatten(items))


def async(func):
    """Poor man's decorator to execute a function asynchronously using separate
    thread. This is an extremely limited async method which doesn't provide any
    way to access the created thread. Use with caution!

    For more robust async, consider using alternative methods/libraries
    (e.g. gevent).
    """
    @wraps(func)
    def wrapper(*args, **kargs):  # pylint: disable=missing-docstring
        thr = Thread(target=func, args=args, kwargs=kargs)
        thr.start()
    return wrapper


def to_dict(data=None, namespace=None):
    """Decorator enabled version of `_to_dict()`."""
    __to_dict = partial(_to_dict, namespace=namespace)

    if data is None:
        # pylint: disable=missing-docstring
        def decorator(func):
            @wraps(func)
            def decorated(*args, **kargs):
                data = func(*args, **kargs)
                return __to_dict(data)
            return decorated
        return decorator
    else:
        return __to_dict(data)


def _to_dict(data, namespace=None):
    """Converts elements of `data` using `data.to_dict()` or
    `data[].to_dict()`. Optionally namespaces data into a container dict using
    `namespace` as key.
    """
    if isinstance(data, list):
        data = [item.to_dict() if hasattr(item, 'to_dict') else item
                for item in data]
    elif hasattr(data, 'to_dict'):
        data = data.to_dict()

    if namespace is not None:
        data = {namespace: data}

    return data


def jsonify(func=None, *args, **kargs):
    """Function or decorator that returns jsonfiy response"""
    if callable(func):
        # pylint: disable=missing-docstring
        @wraps(func)
        def decorated(*args, **kargs):
            return _jsonify(**func(*args, **kargs))
        return decorated
    else:
        if func is not None:
            # consider `f` a positional arg
            args = tuple([func] + list(args))
        return _jsonify(*args, **kargs)



"""Flask extension which integrates additional loggers with Flask app.
"""

import logging
from logging import getLogger, Formatter
from logging.handlers import (
    SMTPHandler as SMTPHandlerBase,
    RotatingFileHandler as RotatingFileHandlerBase
)
import pprint
from functools import partial

from flask import request, current_app


# Universal error message format for custom logger handlers. Any handler who
# uses this should use the EnvironDataMixin class so that extra data
# is attached.
ERROR_FORMAT = """
===============================================================================
Message type:       %(levelname)s
Location:           %(pathname)s:%(lineno)d
Module:             %(module)s
Function:           %(funcName)s
Time:               %(asctime)s


REQUEST:

%(environ)s


ERROR:

%(message)s
===============================================================================
"""

WARNING_FORMAT = (
    '%(asctime)s: %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]')


# default pprint.pformat function to use to output request environ
# pylint: disable=invalid-name
pformat = partial(pprint.pformat, indent=4, depth=None)
# pylint: enable=invalid-name


class Logger(object):
    """Flask Logger extension."""

    def __init__(self, app=None):
        self.app = app

        if app:  # pragma: no cover
            self.init_app(app)

    def init_app(self, app):
        """Initialize app."""
        app.config.setdefault('CARAFE_LOGGER_ENABLED', True)

        if not app.config['CARAFE_LOGGER_ENABLED']:  # pragma: no cover
            return

        if app.config.get('CARAFE_LOGGER_RFILE_ENABLED'):
            loggers = ([app.logger] +
                       app.config.get('CARAFE_LOGGER_RFILE_ADD_LOGGERS', []))

            self.add_handlers(
                loggers, create_rotating_file_handler(app.config))

        if app.config.get('CARAFE_LOGGER_SMTP_ENABLED'):
            loggers = ([app.logger] +
                       app.config.get('CARAFE_LOGGER_SMTP_ADD_LOGGERS', []))

            self.add_handlers(
                loggers, create_email_handler(app.config))

    def add_handlers(self, loggers, handlers):
        """Attach additional logging handlers to loggers."""
        if not isinstance(loggers, list):  # pragma: no cover
            loggers = [loggers]

        if not isinstance(handlers, list):  # pragma: no cover
            handlers = [handlers]

        for logger in loggers:
            if isinstance(logger, basestring):
                logger = getLogger(logger)

            for handler in handlers:
                logger.addHandler(handler)

    def __getattr__(self, attr):
        """Proxy attribute calls to current_app.logger."""
        return getattr(current_app.logger, attr)


class EnvironDataMixin(object):
    """Mixin class that exposes request data to log format."""

    def get_request_environ(self):
        """Return copy of request data for debugging purposes."""
        environ = request.environ.copy()
        environ.update({
            'REQUEST_DATA': request.data,
            'REQUEST_JSON': request.get_json(silent=True),
            'REQUEST_FORM': request.form.to_dict(),
            'REQUEST_FILES': request.files.to_dict(),
            'REQUEST_HEADERS': dict(request.headers),
            'REQUEST_ARGS': request.args.to_dict()
        })

        return environ


class SMTPHandler(SMTPHandlerBase, EnvironDataMixin):
    """Override `emit()` function so that request data is available to log
    formatter.
    """

    def getSubject(self, record):
        """Override with custom subject."""
        return '[{0}] Application Error: {1}'.format(
            current_app.name, request.url)

    def emit(self, record):
        """Override parent method and attach request environ data to
        log record.
        """
        record.environ = pformat(self.get_request_environ())
        super(SMTPHandler, self).emit(record)


class RotatingFileHandler(RotatingFileHandlerBase, EnvironDataMixin):
    """Override `emit()` function so that request data is available to log
    formatter.
    """
    def emit(self, record):
        """Override parent method and attach request environ data to
        log record
        """
        record.environ = pformat(self.get_request_environ())
        super(RotatingFileHandler, self).emit(record)


def create_rotating_file_handler(config):
    """Create a rotating file logger handler."""
    kargs = {
        'filename': config['CARAFE_LOGGER_RFILE_FILENAME'],
        'mode': config.get('CARAFE_LOGGER_RFILE_MODE', 'a'),
        'maxBytes': config.get('CARAFE_LOGGER_RFILE_MAXBYTES', 0),
        'backupCount': config.get('CARAFE_LOGGER_RFILE_BACKUPCOUNT', 0),
        'encoding': config.get('CARAFE_LOGGER_RFILE_ENCODING'),
        'delay': config.get('CARAFE_LOGGER_RFILE_DELAY', 0)
    }

    handler = RotatingFileHandler(**kargs)
    handler.setLevel(
        getattr(logging, config.get('CARAFE_LOGGER_RFILE_LEVEL', 'WARNING')))

    handler.setFormatter(Formatter(WARNING_FORMAT))

    return handler


def create_email_handler(config):
    """Create an email logging handler."""
    if isinstance(config['CARAFE_LOGGER_SMTP_TOADDRS'], basestring):
        # convert to list
        config['CARAFE_LOGGER_SMTP_TOADDRS'] = [
            x.strip()
            for x in config['CARAFE_LOGGER_SMTP_TOADDRS'].split(',')
        ]

    kargs = {
        'mailhost': (
            config['CARAFE_LOGGER_SMTP_SERVER'],
            config['CARAFE_LOGGER_SMTP_PORT']
        ),
        'fromaddr': config['CARAFE_LOGGER_SMTP_FROMADDR'],
        'toaddrs': config['CARAFE_LOGGER_SMTP_TOADDRS'],
        'subject': 'Application Error',
        'credentials': (
            config['CARAFE_LOGGER_SMTP_USERNAME'],
            config['CARAFE_LOGGER_SMTP_PASSWORD']
        ),
        'secure': () if config.get('CARAFE_LOGGER_SMTP_USE_TLS') else None
    }

    handler = SMTPHandler(**kargs)
    handler.setLevel(
        getattr(logging, config.get('CARAFE_LOGGER_SMTP_LEVEL', 'ERROR')))

    handler.setFormatter(Formatter(ERROR_FORMAT))

    return handler



"""Define project metadata
"""

__all__ = [
    '__title__',
    '__summary__',
    '__url__',
    '__version__',
    '__author__',
    '__email__',
    '__license__',
]

__title__ = 'carafe'
__summary__ = 'Collection of Flask extensions geared towards JSON APIs'
__url__ = 'https://github.com/dgilland/carafe'

__version__ = '0.4.0'

__author__ = 'Derrick Gilland'
__email__ = 'dgilland@gmail.com'

__license__ = 'MIT License'



"""Flask extension of Flask-Principal for handling
authentication/authorization.
"""

from flask import session, current_app
from flask_principal import (
    Principal,
    identity_loaded,
    identity_changed,
    Identity,
    AnonymousIdentity,
    Permission,
    RoleNeed,
    TypeNeed
)


# pylint: disable=invalid-name
login_need = TypeNeed('login')
# pylint: enable=invalid-name


class SQLAlchemyAuthProvider(object):
    """SQLAlchemy backed UserProvider class which provides identity information
    for logged in user.
    """

    # SQLAlchemy model used to query the database for identity information
    __model__ = None

    __id_key__ = 'id'
    __roles_key__ = 'roles'

    def __init__(self, db_session):
        self.session = db_session

    def identify(self, identity):
        """Identify a user via _id to provide information for role based
        authentication.
        """
        ident = {}

        user = self.get_user(identity.id)
        if user:
            ident.update({
                self.__id_key__: identity.id,
                self.__roles_key__: self.get_roles(user)
            })

        return ident

    ##
    # Override these methods if they don't return user data correctly based on
    # your models.
    ##

    def get_user(self, _id):
        """Return user object from database."""
        if _id is None:
            return None
        return self.session.query(self.__model__).get(_id)

    def get_roles(self, user):
        """Return a list of `roles` as strings."""
        return getattr(user, self.__roles_key__, [])


class Auth(object):
    """Auth extension."""
    _extension_name = 'carafe.auth'

    def __init__(self, app=None, provider=None):
        self.principal = Principal(use_sessions=False)
        self.require = PermissionFactory()

        self.app = app
        if self.app:  # pragma: no cover
            self.init_app(app, provider)

    def init_app(self, app, provider=None):
        """Initialize app."""
        app.config.setdefault('CARAFE_AUTH_ENABLED', True)
        app.config.setdefault('CARAFE_AUTH_SESSION_ID_KEY', 'user_id')
        app.config.setdefault('CARAFE_AUTH_IDENTITY_ID_KEY', 'id')
        app.config.setdefault('CARAFE_AUTH_IDENTITY_ROLES_KEY', 'roles')

        if not app.config['CARAFE_AUTH_ENABLED']:  # pragma: no cover
            return

        if not hasattr(app, 'extensions'):  # pragma: no cover
            app.extensions = {}

        app.extensions[self._extension_name] = {'provider': provider}

        # NOTE: Instead of having principal use it's session loader, we'll use
        # ours.
        self.principal.init_app(app)
        self.principal.identity_loader(self.session_identity_loader)
        identity_loaded.connect_via(app)(self.on_identity_loaded)

    @property
    def session_id_key(self):
        """Property access to config's CARAFE_AUTH_SESSION_ID_KEY."""
        return current_app.config['CARAFE_AUTH_SESSION_ID_KEY']

    @property
    def identity_id_key(self):
        """Property access to config's CARAFE_AUTH_IDENTITY_ID_KEY."""
        return current_app.config['CARAFE_AUTH_IDENTITY_ID_KEY']

    @property
    def identity_roles_key(self):
        """Property access to config's CARAFE_AUTH_IDENTITY_ROLES_KEY."""
        return current_app.config['CARAFE_AUTH_IDENTITY_ROLES_KEY']

    @property
    def user_id(self):
        """Property access to logged in user id."""
        return session.get(self.session_id_key)

    @property
    def provider(self):
        """Property access to auth provider instance."""
        return current_app.extensions[self._extension_name]['provider']

    def session_identity_loader(self):
        """Fetch user id from session using config's auth id key"""
        if self.session_id_key in session:
            identity = Identity(session[self.session_id_key])
        else:
            identity = None

        return identity

    def on_identity_loaded(self, app, identity):  # pylint: disable=unused-argument
        """Called if session_identity_loader() returns an identity (i.e. not
        None).
        """
        # Whatever is returned is used for our identity. Potentially, provider
        # may return a different user than original identity (e.g. app provides
        # way for admin users to access site using a different user account)

        if self.provider:
            ident = self.provider.identify(identity)
        else:
            ident = {self.identity_id_key: None}

        if self.user_id and not ident:
            # session has an user_id but the ident return is empty
            # user possibly deleted or inactivated in another process
            self.logout()

        # provide auth (whether user is not anonymous)
        if ident.get(self.identity_id_key):
            identity.provides.add(login_need)

        # provide roles
        for role in ident.get(self.identity_roles_key, []):
            identity.provides.add(RoleNeed(role))

    def send_identity_changed(self, user_id):
        """Send identity changed event."""
        if user_id is None:
            identity = AnonymousIdentity()
        else:
            identity = Identity(user_id)

        identity_changed.send(
            current_app._get_current_object(), identity=identity)

    def login(self, user_id, propagate=True):
        """Call after user has been authenticated for login."""
        if session.get(self.session_id_key) != user_id:
            session[self.session_id_key] = user_id
            if propagate:
                self.send_identity_changed(user_id)

    def logout(self, propagate=True):
        """Call to log user out."""
        if session.get(self.session_id_key):
            del session[self.session_id_key]
            if propagate:
                self.send_identity_changed(None)


class PermissionFactory(object):
    """General purpose permissions factory which creates RoleNeed permissions
    from attribute access.
    """
    def __init__(self):
        self._permissions = {
            'login': Permission(login_need)
        }

    def __getattr__(self, role):
        """Return role permission's require method. If it doesn't exist yet,
        create it."""
        if role not in self._permissions:
            self._permissions[role] = Permission(RoleNeed(role))

        return self._permissions[role].require



"""Extension of flask.Response.
"""

from flask import Response as ResponseBase, json, current_app, request


class Response(ResponseBase):
    """Extend flask.Response with support for list/dict conversion to JSON."""
    def __init__(self, content=None, *args, **kargs):
        if isinstance(content, (list, dict)):
            kargs['mimetype'] = 'application/json'
            content = to_json(content)

        super(Response, self).__init__(content, *args, **kargs)

    @classmethod
    def force_type(cls, response, environ=None):
        """Override with support for list/dict."""
        if isinstance(response, (list, dict)):
            return cls(response)
        else:
            return super(Response, cls).force_type(response, environ)


def to_json(content):
    """Converts content to json while respecting config options."""
    indent = None
    if (current_app.config['JSONIFY_PRETTYPRINT_REGULAR']
            and not request.is_xhr):
        indent = 2
    return json.dumps(content, indent=indent)



"""Collection of Flask extensions geared towards JSON APIs.
"""

from .app import FlaskCarafe
from .client import Client, JSONClient

from .__meta__ import (
    __title__,
    __summary__,
    __url__,
    __version__,
    __author__,
    __email__,
    __license__
)



"""Carafe's custom Flask app.
"""

from flask import Flask

from .request import Request
from .response import Response


class FlaskCarafe(Flask):
    """Extension of standard Flask app with custom request and response
    classes.
    """
    request_class = Request
    response_class = Response



"""Flask extension of Flask-Cache.
"""

from functools import wraps

from flask import request, current_app
from flask.signals import Namespace
from werkzeug import urls
from flask_cache import Cache as CacheBase


# pylint: disable=invalid-name


signals = Namespace()


# Signals for dealing with cache invalidation after REST methods.
after_post = signals.signal('after_post', doc="""
Signal which should be sent after a POST operation.
""")


after_put = signals.signal('after_put', doc="""
Signal which should be sent after a PUT operation.
""")


after_patch = signals.signal('after_patch', doc="""
Signal which should be sent after a PATCH operation.
""")


after_delete = signals.signal('after_delete', doc="""
Signal which should be sent after a DELETE operation.
""")


# pylint: enable=invalid-name


class Cache(CacheBase):
    """Manager class for handling creating and deleting cache keys based o
    view events.
    """

    view_key_format = '{namespace}:view:{path}'

    def init_app(self, app, config=None):
        if config is None:
            config = app.config

        config.setdefault('CARAFE_CACHE_ENABLED', True)
        config.setdefault('CARAFE_CACHE_IGNORED_REQUEST_ARGS', [])

        if not config['CARAFE_CACHE_ENABLED']:  # pragma: no cover
            return

        super(Cache, self).init_app(app, config=config)

        self.connect_signals()

    def connect_signals(self):
        """Connect supported signals to handlers."""
        after_post.connect(self.on_after_post)
        after_put.connect(self.on_after_put)
        after_patch.connect(self.on_after_patch)
        after_delete.connect(self.on_after_delete)

    def get_cache_namespace(self, obj):
        """Determine object's cache namespace."""
        if getattr(obj, 'cache_namespace', None) is not None:
            return obj.cache_namespace
        elif hasattr(obj, '__name__'):
            return obj.__name__
        else:
            return obj.__class__.__name__

    @property
    def client(self):
        """Proxy to cache client wrapper."""
        return self.cache if self.enabled else None

    @property
    def server(self):
        """Proxy to cache server client."""
        return getattr(self.cache, '_client', None) if self.enabled else None

    @property
    def enabled(self):
        """Property access to config's CARAFE_CACHE_ENABLED."""
        return current_app.config['CARAFE_CACHE_ENABLED']

    def clear_keys(self, pipe, *keys, **kargs):
        """Clear specified keys"""
        if not keys:
            return

        execute = kargs.get('execute')

        keys = [current_app.config['CACHE_KEY_PREFIX'] + k for k in keys]
        pipe.delete(*keys)

        if execute:
            pipe.execute()

        return True

    def clear_prefix(self, pipe, *prefixes, **kargs):
        """Clear keys starting with prefix"""
        keys = []
        for prefix in prefixes:
            keys.extend(pipe.keys(prefix + '*'))

        return self.clear_keys(pipe, *keys, **kargs)

    def clear(self, prefixes=None, keys=None):
        """Clear cache keys using an optional prefix, regex, and/or list of
        keys.
        """
        if not self.enabled:
            return

        if not any([prefixes, keys]):
            # this is the same as clearing the entire cache
            return self.cache.clear()

        if not hasattr(self.server, 'pipeline'):
            # enhanced cache clearing is only supported using redis
            return self.cache.clear()

        if prefixes is None:
            prefixes = []

        if keys is None:
            keys = []

        with self.server.pipeline() as pipe:
            self.clear_prefix(pipe, *prefixes)
            self.clear_keys(pipe, *keys)
            pipe.execute()

        return True

    def cached_view(self,
                    timeout=None,
                    namespace=None,
                    unless=None,
                    include_request_args=True):
        """Wrapper around self.cached which itself is a decorator. We're
        wrapping because we want to have access to the class instance of the
        view in order to namespace the key. We can't always namespace using
        key_prefix since some cache decorators are placed around parent classes
        which don't know anything about the child class.
        """

        # pylint: disable=missing-docstring
        def wrap(func):
            @wraps(func)
            def wrapper(*args, **kargs):
                if not self.enabled:
                    return func(*args, **kargs)

                if namespace is not None:
                    # Make namespace available in case `f` is used as signal
                    # sender. Mainly used to get namespace when invalidating
                    # cache keys via namespace prefix.
                    func.cache_namespace = namespace

                # If args[0] is set, then this is a class based view, else use
                # function.
                obj = args[0] if args else func
                cache_namespace = self.get_cache_namespace(obj)
                view_path = self.create_view_path(include_request_args)
                key_prefix = self.view_key_format.format(
                    namespace=cache_namespace,
                    path=view_path,
                    **request.view_args)

                cached = self.cached(timeout=timeout,
                                     key_prefix=key_prefix,
                                     unless=unless)(func)

                try:
                    # Cache server could be down.
                    result = cached(*args, **kargs)
                except Exception as ex:  # pragma: no cover
                    # Return function call instead.
                    current_app.logger.exception(ex)
                    result = func(*args, **kargs)

                return result

            return wrapper

        return wrap

    def create_view_path(self, include_request_args=False):
        """Construct view path from request.path with option to include GET
        args.
        """
        href = urls.Href(request.path)

        if include_request_args:
            ignored = current_app.config['CARAFE_CACHE_IGNORED_REQUEST_ARGS']
            args = dict((k, v) for k, v in request.args.iteritems()
                        if k not in ignored)
        else:
            args = None

        return href(args)

    def on_modified_record(self, sender):
        """Common tasks to perform when a record is modified."""
        namespace = self.get_cache_namespace(sender)
        prefixes = [self.view_key_format.format(namespace=namespace, path='')]

        # Append cascade keys which should be invalidated (typically due to
        # this API's data being used in other APIs).
        prefixes += getattr(sender, 'cache_cascade', [])

        try:
            self.clear(prefixes=prefixes)
        except Exception as ex:  # pragma: no cover
            current_app.logger.exception(ex)

    def on_after_post(self, sender):
        """Handle the `after_post` event. Executed after a POST request."""
        self.on_modified_record(sender)

    def on_after_put(self, sender):
        """Handle the `after_put` event. Executed after a PUT request."""
        self.on_modified_record(sender)

    def on_after_patch(self, sender):
        """Handle the `after_patch` event. Executed after a PATCH request."""
        self.on_modified_record(sender)

    def on_after_delete(self, sender):
        """Handle the `after_delete` event. Executed after a DELETE request."""
        self.on_modified_record(sender)



"""
carafe
======

Flask application factory with extensions geared towards JSON APIs.

Documentation: https://github.com/dgilland/carafe
"""

from setuptools import setup, find_packages


meta = {}
with open('carafe/__meta__.py') as fp:
    exec(fp.read(), meta)


setup(
    name=meta['__title__'],
    version=meta['__version__'],
    url=meta['__url__'],
    license=meta['__license__'],
    author=meta['__author__'],
    author_email=meta['__email__'],
    description=meta['__summary__'],
    long_description=__doc__,
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
    install_requires=[
        'Flask>=0.10.1',
        'Flask-Cache>=0.12',
        'Flask-Principal>=0.4.0',
        'Flask-Testing>=0.4',
        'blinker>=1.3',
    ],
    test_suite='tests',
    keywords='flask',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Framework :: Flask',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ]
)
